/* WARNING: This code is just a quick hack based on some source I wrote for
   the FreeBASIC DOS gfxlib.  It is not meant to be clean or well-written or
   nice to look at...

   Use at your own risk.
*/

/*
 vbediag version 0.1
 Copyright (C) 2007 Daniel R. Verkamp

 This software is provided 'as-is', without any express or implied warranty. 
 In no event will the authors be held liable for any damages arising from 
 the use of this software.
 
 Permission is granted to anyone to use this software for any purpose, 
 including commercial applications, and to alter it and redistribute it 
 freely, subject to the following restrictions:
 
 1. The origin of this software must not be misrepresented; you must not claim 
 that you wrote the original software. If you use this software in a product, 
 an acknowledgment in the product documentation would be appreciated but is 
 not required.
 
 2. Altered source versions must be plainly marked as such, and must not be 
 misrepresented as being the original software.
 
 3. This notice may not be removed or altered from any source 
 distribution. 
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/farptr.h>
#include <pc.h>
#include <dos.h>
#include <dpmi.h>
#include <go32.h>

#define TRUE 1
#define FALSE 0

#define MASK_LINEAR(addr)   ((addr) & 0x000FFFFF)
#define RM_TO_LINEAR(addr)  ((((addr) & 0xFFFF0000) >> 12) + ((addr) & 0xFFFF))
#define RM_OFFSET(addr)     ((addr) & 0xF)
#define RM_SEGMENT(addr)    (((addr) >> 4) & 0xFFFF)

/* Info block for the VESA controller */
typedef struct VbeInfoBlock
{
	char            vbe_signature[4];	/* Must be 'VESA', or 'VBE2' for VESA2          */
	unsigned short  vbe_version;		/* 0200h for VESA2                              */
	unsigned int    oem_string_ptr;		/* Pointer to OEM string                        */
	unsigned int    capabilities;		/* Capabilities of the controller               */
	unsigned int    video_mode_ptr;		/* Pointer to a video mode list                 */
	unsigned short  total_memory;		/* Number of 64 KB memory blocks (VESA2)        */
	unsigned short  oem_software_rev;	/* VBE implementation software revision         */
	unsigned int    oem_vendor_name;	/* Pointer to Vendor Name string                */
	unsigned int    oem_product_name;	/* Pointer to Product Name string               */
	unsigned int    oem_product_rev;	/* Pointer to Product Revision string           */
	unsigned char   reserved[222];		/* Reserved for VBE implementation scratch area */
	unsigned char   oem_data[256];		/* Data area for OEM strings                    */
}
__attribute__ ((packed)) VbeInfoBlock;


/* Mnemonics for VbeInfoBlock.capabilities flags */
enum
{
	VIB_8BITDAC = 1 << 0, /* DAC width switchable to 8 bits per primary color */
	VIB_NOTVGA  = 1 << 1, /* Controller is not VGA compatible                 */
	VIB_BLANK   = 1 << 2  /* RAMDAC recommentds programming during blank      */
};


/* Info block for a VESA video mode */
typedef struct VesaModeInfo
{
	/* mandatory for all VBE versions */
	unsigned short ModeAttributes;
	unsigned char  WinAAttributes;
	unsigned char  WinBAttributes;
	unsigned short WinGranularity;
	unsigned short WinSize;
	unsigned short WinASegment;
	unsigned short WinBSegment;
	unsigned long  WinFuncPtr;
	unsigned short BytesPerScanLine;
	
	/* mandatory for VBE 1.2 and above */
	unsigned short XResolution;
	unsigned short YResolution;
	unsigned char  XCharSize;
	unsigned char  YCharSize;
	unsigned char  NumberOfPlanes;
	unsigned char  BitsPerPixel;
	unsigned char  NumberOfBanks;
	unsigned char  MemoryModel;
	unsigned char  BankSize;              /* bank size in KB */
	unsigned char  NumberOfImagePages;
	unsigned char  Reserved_page;
	
	/* direct color fields (required for direct and YUV memory models) */
	unsigned char  RedMaskSize;
	unsigned char  RedFieldPosition;
	unsigned char  GreenMaskSize;
	unsigned char  GreenFieldPosition;
	unsigned char  BlueMaskSize;
	unsigned char  BlueFieldPosition;
	unsigned char  RsvdMaskSize;
	unsigned char  RsvdFieldPosition;
	unsigned char  DirectColorModeInfo;
	
	/* mandatory for VBE 2.0 and above */
	unsigned long  PhysBasePtr;
	unsigned long  OffScreenMemOffset;    /* reserved in VBE 3.0 */
	unsigned short OffScreenMemSize;      /* reserved in VBE 3.0 */
	
	/* mandatory for VBE 3.0 and above */
	unsigned short LinBytesPerScanLine;
	unsigned char  BnkNumberOfImagePages;
	unsigned char  LinNumberOfImagePages;
	unsigned char  LinRedMaskSize;
	unsigned char  LinRedFieldPosition;
	unsigned char  LinGreenMaskSize;
	unsigned char  LinGreenFieldPosition;
	unsigned char  LinBlueMaskSize;
	unsigned char  LinBlueFieldPosition;
	unsigned char  LinRsvdMaskSize;
	unsigned char  LinRsvdFieldPosition;
	unsigned long  MaxPixelClock;
	
	unsigned char  Reserved[189];
	
} __attribute__ ((packed)) VesaModeInfo;


/* Mnemonics for VesaModeInfo.ModeAttributes flags */
enum
{
	VMI_MA_SUPPORTED = 1 << 0,  /* Mode supported by hardware configuration   */
	VMI_MA_TTY       = 1 << 2,  /* TTY output functions supported by the BIOS */
	VMI_MA_COLOR     = 1 << 3,  /* Color mode if set, monochrome if not       */
	VMI_MA_GRAPHICS  = 1 << 4,  /* Graphics mode if set, text mode if not     */
	VMI_MA_NOTVGA    = 1 << 5,  /* VGA compatible mode                        */
	VMI_MA_NOTVGAWIN = 1 << 6,  /* VGA compatible windowed memory available   */
	VMI_MA_LFB       = 1 << 7,  /* Linear frame buffer mode available         */
	VMI_MA_DBLSCAN   = 1 << 8,  /* Double scan mode is available              */
	VMI_MA_INTERLACE = 1 << 9,  /* Interlaced mode is available               */
	VMI_MA_TRIPLEBUF = 1 << 10, /* Hardware triple buffering support          */
	VMI_MA_STEREO    = 1 << 11, /* Hardware stereoscopic display support      */
	VMI_MA_DUALDISP  = 1 << 12  /* Dual display start address support         */
};


/* Mnemonics for VesaModeInfo.window_attributes flags */
enum
{
	VMI_WA_RELOC = 1 << 0, /* Relocatable window(s) supported */
	VMI_WA_READ  = 1 << 1, /* Window is readable              */
	VMI_WA_WRITE = 1 << 2  /* Window is writeable             */
};


/* Valid values for VesaModeInfo.mem_model */
typedef enum VmiMemModel
{
	VMI_MM_TEXT = 0x00, /* Text mode                   */
	VMI_MM_CGA  = 0x01, /* CGA graphics                */
	VMI_MM_HERC = 0x02, /* Hercules graphics           */
	VMI_MM_PLAN = 0x03, /* Planar                      */
	VMI_MM_PACK = 0x04, /* Packed pixel                */
	VMI_MM_NCHN = 0x05, /* Non-chain 4, 256 color      */
	VMI_MM_DIR  = 0x06, /* Direct color                */
	VMI_MM_YUV  = 0x07, /* YUV color                   */
	VMI_MM_OEM  = 0x10  /* Start of OEM defined models */
}
VmiMemModel;


/* Mnemonics for VesaModeInfo.direct_color_mode_info */
enum
{
	VMI_DC_RAMP = 1 << 0, /* Color ramp programmable if set, fixed if not     */
	VMI_DC_RSVD = 1 << 1  /* Bits in rsvd field are usable by the application */
};

typedef struct CRTCInfoBlock
{
	unsigned short HorizontalTotal;     /* horizontal total in pixels */
	unsigned short HorizontalSyncStart; /* horizontal sync start in pixels */
	unsigned short HorizontalSyncEnd;   /* horizontal sync end in pixels */
	unsigned short VerticalTotal;       /* vertical total in lines */
	unsigned short VerticalSyncStart;   /* vertical sync start in lines */
	unsigned short VerticalSyncEnd;     /* vertical sync end in lines */
	unsigned char  Flags;               /* flags (interlaced, double scan, etc.) */
	unsigned long  PixelClock;          /* pixel clock in units of Hz */
	unsigned short RefreshRate;         /* refresh rate in units of 0.01 Hz */
	
	unsigned char  Reserved[40];
} __attribute__ ((packed)) CRTCInfoBlock;

/* flags for CRTCInfoBlock.Flags */
enum
{
	CRTC_DBLSCAN   = 1 << 0, /* double scan mode enable */
	CRTC_INTERLACE = 1 << 1, /* interlaced mode enable */
	CRTC_HORZSYNC  = 1 << 2, /* horizontal sync polarity (clear = +, set = -) */
	CRTC_VERTSYNC  = 1 << 3  /* vertical sync polarity (clear = +, set = -) */
};

typedef struct VesaPMInfo
{
	unsigned short setWindow;
	unsigned short setDisplayStart;
	unsigned short setPalette;
	unsigned short IOPrivInfo;
} __attribute__ ((packed)) VesaPMInfo;

VbeInfoBlock vesa_info;
VesaModeInfo vesa_mode_info;
__dpmi_regs regs;
int vesa_ok;
int num_vesa_modes;

static int vesa_get_mode_info(int mode)
{
	int i;
	
	memset(&vesa_mode_info, 0, sizeof(vesa_mode_info));

	_farsetsel(_dos_ds);

	for (i = 0; i < sizeof(vesa_mode_info); i++) {
		_farnspokeb(MASK_LINEAR(__tb) + i, 0);
	}

	regs.x.ax = 0x4F01;
	regs.x.di = RM_OFFSET(__tb);
	regs.x.es = RM_SEGMENT(__tb);
	regs.x.cx = mode;
	__dpmi_int(0x10, &regs);
	if (regs.h.ah)
		return -1;

	dosmemget(MASK_LINEAR(__tb), sizeof(vesa_mode_info), &vesa_mode_info);
	return 0;
}

static const char *get_far_string(long ptr)
{
	static char buffer[1024];
	char *s = buffer;
	
	if (!ptr)
	{
		buffer[0] = '\0';
	}
	else
	{
		_farsetsel(_dos_ds);
		while (*(s++) = _farnspeekb(ptr++));
	}
	return buffer;
}

void fb_dos_vesa_detect(void)
{
	int i;
	int mode_list[256];
	int number_of_modes;
	long mode_ptr;
	int c;
	int attribs = VMI_MA_SUPPORTED | VMI_MA_COLOR | VMI_MA_GRAPHICS;
	
	printf("== VBEDIAG v0.1 ==\n");
		
	/* detect VESA */
	
	vesa_ok = FALSE;
	
	_farsetsel(_dos_ds);

	for (i = 4; i < (int)sizeof(VbeInfoBlock); i++)
		_farnspokeb(MASK_LINEAR(__tb) + i, 0);
	
	dosmemput("VBE2", 4, MASK_LINEAR(__tb)); /* get VESA 2 info if available */
	
	regs.x.ax = 0x4F00;
	regs.x.di = RM_OFFSET(__tb);
	regs.x.es = RM_SEGMENT(__tb);
	__dpmi_int(0x10, &regs);
	
	if (regs.h.ah == 0x00)
	{
		dosmemget(MASK_LINEAR(__tb), sizeof(VbeInfoBlock), &vesa_info);
		if (strncmp(vesa_info.vbe_signature, "VESA", 4) == 0)
			vesa_ok = TRUE;
	}
	
	/* get VESA modes */
	if (vesa_ok)
	{
		printf("VESA VBE version %d.%d%d\n", (vesa_info.vbe_version >> 8) & 0xF, (vesa_info.vbe_version >> 4) & 0xF, vesa_info.vbe_version & 0xF);
		printf("OEM: '%s'\n", get_far_string(RM_TO_LINEAR(vesa_info.oem_string_ptr)));
		
		if (vesa_info.vbe_version >= 0x200)
		{
			printf("Capabilities: %08X = ", vesa_info.capabilities);
			if (vesa_info.capabilities & VIB_8BITDAC) printf("8bitdac ");
			if (vesa_info.capabilities & VIB_NOTVGA)  printf("notvga ");
			if (vesa_info.capabilities & VIB_BLANK)   printf("ramdacblank ");
			printf("\n");
			printf("Total memory: %d bytes = %d KiB = %d MiB\n", (unsigned int)vesa_info.total_memory << 16, (unsigned int)vesa_info.total_memory << 6, (unsigned int)vesa_info.total_memory >> 4);
			printf("OEM software rev: %d%d.%d%d\n", (vesa_info.oem_software_rev >> 12) & 0xF, (vesa_info.oem_software_rev >> 8) & 0xF, (vesa_info.oem_software_rev >> 4) & 0xF, vesa_info.oem_software_rev & 0xF);
			printf("OEM vendor name: '%s'\n", get_far_string(RM_TO_LINEAR(vesa_info.oem_vendor_name)));
			printf("OEM product name: '%s'\n", get_far_string(RM_TO_LINEAR(vesa_info.oem_product_name)));
			printf("OEM product rev: '%s'\n", get_far_string(RM_TO_LINEAR(vesa_info.oem_product_rev)));
		}
		
		mode_ptr = RM_TO_LINEAR(vesa_info.video_mode_ptr);
		
		number_of_modes = 0;
		
		while (_farpeekw(_dos_ds, mode_ptr) != 0xFFFF)
		{
			mode_list[number_of_modes] = _farpeekw(_dos_ds, mode_ptr);
			number_of_modes++;
			mode_ptr += 2;
		}
		
		num_vesa_modes = number_of_modes;
		
		printf("Number of modes: %d\n", num_vesa_modes);
		
		for (c = 0; c < number_of_modes; c++)
		{
			printf("mode #%d [%03X]: ", c, mode_list[c]);
			if ( (vesa_get_mode_info(mode_list[c]) == 0) )
			{
				printf("%dx%dx%d bpp", vesa_mode_info.XResolution, vesa_mode_info.YResolution, vesa_mode_info.BitsPerPixel);
				printf(", %d plane(s), %d bank(s) of %d KiB, %d B/line,", vesa_mode_info.NumberOfPlanes, vesa_mode_info.NumberOfBanks, vesa_mode_info.BankSize, vesa_mode_info.BytesPerScanLine);
				printf("\n\t");
				if ((vesa_mode_info.WinAAttributes | vesa_mode_info.WinBAttributes) & 1)
				{
					printf("wingran=%d KiB, winsize=%d KiB", vesa_mode_info.WinGranularity, vesa_mode_info.WinSize);
				}
				if (vesa_mode_info.ModeAttributes & VMI_MA_SUPPORTED) printf(", supported");
				if (vesa_mode_info.ModeAttributes & VMI_MA_TTY) printf(", tty");
				if (vesa_mode_info.ModeAttributes & VMI_MA_COLOR) printf(", color");
				if (vesa_mode_info.ModeAttributes & VMI_MA_GRAPHICS) printf(", gfx");
				if (vesa_mode_info.ModeAttributes & VMI_MA_NOTVGA) printf(", notvga");
				if (vesa_mode_info.ModeAttributes & VMI_MA_NOTVGAWIN) printf(", notvgawin");
				if (vesa_mode_info.ModeAttributes & VMI_MA_LFB) printf(", lfb");
				printf("\n\t");
				if (vesa_mode_info.WinAAttributes)
				{
					printf("winA:");
					if (vesa_mode_info.WinAAttributes & VMI_WA_RELOC) printf(" reloc");
					if (vesa_mode_info.WinAAttributes & VMI_WA_RELOC) printf(" read");
					if (vesa_mode_info.WinAAttributes & VMI_WA_WRITE) printf(" write");
				}

				if (vesa_mode_info.WinBAttributes)
				{
					printf("winB:");
					if (vesa_mode_info.WinBAttributes & VMI_WA_RELOC) printf(" reloc");
					if (vesa_mode_info.WinBAttributes & VMI_WA_RELOC) printf(" read");
					if (vesa_mode_info.WinBAttributes & VMI_WA_WRITE) printf(" write");
				}
				printf("\n\t");
				if (vesa_mode_info.BitsPerPixel > 8)
				{
					printf("R %d(%d) G %d(%d) B %d(%d)", vesa_mode_info.RedMaskSize, vesa_mode_info.RedFieldPosition, vesa_mode_info.GreenMaskSize, vesa_mode_info.GreenFieldPosition, vesa_mode_info.BlueMaskSize, vesa_mode_info.BlueFieldPosition);
					printf("\n\t");
				}

				if  ( ((vesa_mode_info.ModeAttributes & attribs) == attribs) && /* color graphics mode and supported */
				     (vesa_mode_info.NumberOfPlanes == 1) &&
				     ((vesa_mode_info.MemoryModel == VMI_MM_PACK) ||
				      (vesa_mode_info.MemoryModel == VMI_MM_DIR)) &&
					 ((vesa_info.vbe_version < 0x200) ||
					  ((unsigned int)vesa_info.total_memory << 16 >=
					   vesa_mode_info.BytesPerScanLine * vesa_mode_info.YResolution)) )
				{
					printf("usable");
					printf("\n\t");
				}
			}
			
			printf("\n");
		}
	}
	else
	{
		printf("No VESA VBE available!\n");
	}
}

int main(int argc, char **argv)
{
	fb_dos_vesa_detect();
}
